from diagrams.programming.framework import Django

from __init__ import get_conf, adjust_image
from icons import Composer, Manifest, PackageInstaller, Processor, ProjectPart


if __name__ == "__main__":
    from pathlib import Path

    from diagrams import Cluster, Diagram, Edge

    conf = get_conf(Path(__file__).stem)
    output_filepath = conf.pop("output_filepath")

    with Diagram("", **conf):
        manifest = Manifest("Manifest")
        composer = Composer("Composer")

        with Cluster("Processors"):
            settings_proc = Processor("DjangoSettings")
            urls_proc = Processor("DjangoUrls")
            requirements_proc = Processor("TextContent")

            processors = [
                settings_proc,
                urls_proc,
                requirements_proc
            ]

        with Cluster("Django parts"):
            settings_part = ProjectPart("Settings")
            urls_part = ProjectPart("Urls")
            requirements_part = ProjectPart("Requirements")

            parts = [
                settings_part,
                urls_part,
                requirements_part
            ]

        settings_proc >> Edge(color="darkgreen") >> settings_part
        urls_proc >> Edge(color="darkgreen") >> urls_part
        requirements_proc >> Edge(color="chocolate2") >> requirements_part

        django = Django("Django")
        pip = PackageInstaller("Pip")

        settings_part >> Edge(color="darkgreen") >> django
        urls_part >> Edge(color="darkgreen") >> django
        requirements_part >> Edge(color="chocolate2") >> pip

        manifest >> composer

        composer >> Edge(color="darkgreen") >> settings_proc
        composer >> Edge(color="darkgreen") >> urls_proc
        composer >> Edge(color="chocolate2") >> requirements_proc

    adjust_image(output_filepath.resolve())
    print("Built diagram at: {}".format(output_filepath.resolve()))




class AppNode:
    """
    Class to store application payload as an object.

    Arguments:
        name (string): The application module name.

    Keyword Arguments:
        push_end (boolean): Application parameter to describe that it should be pushed
            to end of the ordered application list.

    Attributes:
        name (string): The application name comes from its directory name in
            application repository.
        dependencies (list): List of AppNode objects. Only filled once ``AppStore``
            resolving has been done.
        dependency_names (list): List of string names. It is mostly used in
            ``AppStore`` preparation until it has done resolving.
        push_end (boolean): "Push end" mode value.
    """
    def __init__(self, name, push_end=False):
        self.name = name
        self.dependencies = []
        self.dependency_names = []
        self.push_end = push_end

    def __repr__(self):
        return "<{klass}: {name}>".format(
            klass=self.__class__.__name__,
            name=self.name
        )

    def __str__(self):
        return self.name

    def add_dependency(self, node):
        """
        Add a dependency object to the object registry ``AppNode.dependencies``.

        A dependency which already exists in registry is not added twice.

        Arguments:
            node (AppNode): Dependency object to add.
        """
        if node.name not in [item.name for item in self.dependencies]:
            self.dependencies.append(node)

    def add_dependency_name(self, name):
        """
        Add a dependency name to the name registry ``AppNode.dependency_names``.

        A dependency name which already exists in registry is not added twice.

        Arguments:
            node (string): Dependency name to add.
        """
        if name not in self.dependency_names:
            self.dependency_names.append(name)

    def to_dict(self, flat=False):
        """
        Serialize the object attribute as a dictionnary.

        Keyword Arguments:
            flat (boolean): If set to True, the dependencies are returned as a list of
                names (string) instead of AppNode objects.

        Returns:
            dict: Application payload including dependencies as ``AppNode`` objects.
        """
        return {
            "name": self.name,
            "dependencies": [
                dependency.name if flat is True else dependency
                for dependency in self.dependencies
            ],
            "push_end": self.push_end,
        }

    def to_payload(self):
        """
        Shortcut to ``to_dict`` with flat mode enforced.

        Returns:
            dict: Application payload including dependencies as string name.
        """
        return self.to_dict(flat=True)



"""
AppStore is in charge to store application collection, process it to translate
apps as AppNode objects and resolve application order following criterias.

Order criterias are:

* Natural list order which is a soft ordering because it is only respected after other
  criterias;
* Dependency, basically an application occurs after its dependencies;
* Parameter ``push_end`` to push an application after other applications that don't set
  it to True;

Parameter ``push_end`` is the strongest criterias but it does not break the dependency
ordering, so an app follow its dependencies if they are in `push_end`` mode.

Commonly, the `push_end` is to reserve to some very special applications and you
should prefer to lean on natural order and dependency order. If you use it, it is
recommended to use it the less possible because it may prooduces a cascade of
re-ordering you won't be able to fully control.

This is inspired by the algorith from "Ferry Boender":

https://www.electricmonk.nl/docs/dependency_resolving_algorithm/

"""
from ..exceptions import ComposerAppStoreError

from .node import AppNode


class AppStore:
    """
    Store a collection of applications and manage their dependencies.

    Application collection is a list of dictionnaries with the following structure: ::

        [
            {
                "name": String,
                "push_end": Boolean,
                "dependencies": List[String..]
            },
        ]

    Internally, every application are stored as an ``AppNode`` object.

    Keyword Arguments:
        default_app (string): Application name to attach as dependencies for
            applications that don't have any dependency. The name must exists in given
            collection. By default no default dependency is applied.

    Attributes:
        default_app (string): The value of ``default_app`` argument.
        processed_apps (list): Internal list of processed applications (translated to
            AppNode) filled by ``AppStore.process_collection()``.

    """
    def __init__(self, default_app=None):
        self.default_app = default_app
        self.processed_apps = []

    def get_app(self, name, default=None):
        """
        Get an app object from processed app list.

        Arguments:
            name (string): The name to get from processed applications.

        Keyword Arguments:
            default (object): Default value to use when given name is not retrieved
                from processed applications.

        Returns:
            AppNode: Application object.
        """
        return next(
            (app for app in self.processed_apps if app.name == name),
            default
        )

    def process_collection(self, collection):
        """
        Correctly store a collection of apps.

        This must be called before "resolve" since it register app nodes before linking
        their node dependencies.

        Also this is linear workflow only, at this point the list is not safe for
        circular references.

        Arguments:
            collection (list): List of application datas. Each app data must have a non
                empty ``name`` item. Also accept an optional item ``dependencies``
                which is a list of dependency names and an optional ``push_end`` item.
        """
        # First collect every app with their parameters as an Appnode in registry
        # At this stage app dependencies are only stored as name strings since not all
        # dependencies are yet registered as AppNode
        for item in collection:
            if self.get_app(item.get("name")):
                msg = (
                    "Application '{}' have multiple references in collection."
                )
                raise ComposerAppStoreError(msg.format(item.get("name")))

            node = AppNode(
                item.get("name"),
                push_end=item.get("push_end", False),
            )

            for name in item.get("dependencies", []):
                node.add_dependency_name(name)

            # Append the default dependency when app does not have any
            if (
                self.default_app and
                len(node.dependency_names) == 0 and
                self.default_app != node.name and
                self.default_app not in node.dependency_names
            ):
                node.add_dependency_name(self.default_app)

            self.processed_apps.append(node)

        # Then walk in processed apps to translate their dependency names with
        # registered AppNode
        for app in self.processed_apps:
            for name in app.dependency_names:
                node = self.get_app(name)

                if node:
                    app.add_dependency(node)
                else:
                    msg = (
                        "Dependency '{dep}' from application '{app}' is not a "
                        "registered application."
                    )
                    raise ComposerAppStoreError(msg.format(dep=name, app=app))

    def _apply_recursing_inheritance(self, app):
        """
        Apply dependencies 'push_end' inheritage.

        This will recursively walk into app's dependencies to search for any dependency
        app with 'push_end' to True and if true apply it on app.

        Arguments:
            app (AppNode): Application object to dig in for its dependency inheritance.
        """
        # Apply inheritance if any dep has push_end to True
        if len([
            item.name
            for item in app.dependencies
            if item.push_end
        ]) > 0:
            app.push_end = True

        # Follow dependencies to continue inheritance discovery
        for dep in app.dependencies:
            self._apply_recursing_inheritance(dep)

    def dependency_resolver(self, node, resolved, unresolved):
        """
        Recursive dependency resolver.

        This follow application dependencies to position them in the resolved list
        such a dependency is always after the application which require it.

        Arguments:
            node (AppNode): Application object to walk in.
            resolved (list): List of resolved application objects. Updated during
                resolving. This is commonly the value you will use to get the resolved
                and ordered applications.
            unresolved (list): List of unresolved application objects. Updated during
                resolving.
        """
        unresolved.append(node)

        for dependency in node.dependencies:
            if dependency not in resolved:
                if dependency in unresolved:
                    msg = "Circular reference detected: {source} -> {to}"
                    raise ComposerAppStoreError(
                        msg.format(source=node.name, to=dependency.name)
                    )
                self.dependency_resolver(dependency, resolved, unresolved)

        resolved.append(node)
        unresolved.remove(node)

    def resolve(self, collection, flat=False, no_ordering=False):
        """
        Resolve app list in order of app dependencies such as an app is always after
        all its dependencies.

        Arguments:
            collection (list): List of application payloads to work on.

        Keyword Arguments:
            flat (boolean): If True, returned list will be AppNode payload. Default to
                False.
            no_ordering (boolean): If ``True`` the ``AppStore.resolve()`` directly
                return the AppNode list with original order from collection. There
                won't be any resolution. Default is ``False``.

        Returns:
            list: List of AppNode object or payload (dict) respectively depending flat
            mode is False or True.
        """
        resolved = []

        # Process given application collection to translate them to AppNode with their
        # right parameters
        self.process_collection(collection)

        # By pass further resolving to return the app list ordered with its natural
        # order
        if no_ordering:
            ordered_resolve = self.processed_apps
        # Proceed to the last resolving actions
        else:
            # Go recursively resolve apps order with implied order by dependency
            for node in self.processed_apps:
                if node.name not in [r.name for r in resolved]:
                    self.dependency_resolver(node, resolved, [])

            # Apply possible dependencies parameters inheritance
            for app in resolved:
                self._apply_recursing_inheritance(app)

            # Consume resolved list to distinct apps with push_end=False from those
            # with push=True, built two distinct lists that are then joined (False
            # first, True last).
            ordered_resolve = [
                item
                for item in resolved
                if item.push_end is False
            ] + [
                item
                for item in resolved
                if item.push_end is True
            ]

        if flat:
            return [item.name for item in ordered_resolve]
        else:
            return ordered_resolve



import datetime
import json
from pathlib import Path

from ..app_storage import AppNode


class ExtendedJsonEncoder(json.JSONEncoder):
    """
    Extended JSON encoder to support more basic and internal object types.

    Sample usage: ::

        import json
        json.dumps({}, indent=4, cls=ExtendedJsonEncoder)

    """
    def default(self, obj):
        # Support for pathlib.Path to a string
        if isinstance(obj, Path):
            return str(obj)
        # Support for set to a list
        if isinstance(obj, set):
            return list(obj)
        # Support for AppNode to a dict
        if isinstance(obj, AppNode):
            return obj.to_payload()
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)



import json
from pathlib import Path

import tomli

from ..exceptions import ComposerManifestError

from .base import BaseConfig
from .fields import CharField, ListField, PluginField, BooleanField
from .plugins import RequirementsConfig


class Manifest(BaseConfig):
    """
    The manifest model.

    Manifest fields are given as keyword arguments.

    Fields are:

    name (string)
        The manifest title name.

    collection (list)
        A list of application module names for enabled application.
    repository (string)
        A Python path where to search for enabled application
        modules.
    syspaths (list)
        A list of Path object to load in sys.path by Composer.
    requirements (dict)
        A dictionnary of items to load in ``RequirementsConfig``
        for specific requirements composer. In fact this is used by
        ``TextContentComposer`` but requirements is actually its unique
        implementation.

    Attributes:
        name (string): The manifest title name.
        collection (list): A list of application module names for enabled application.
        requirements (dict or RequirementsConfig): Requirements specific options.
            Either as RequirementsConfig object or a dict of values respecting the
            RequirementsConfig attributes.
    """
    # Payload fields declaration
    _FIELDS = [
        CharField("name", required=True),
        ListField("collection", required=True),
        CharField("repository", required=True),
        CharField("default_store_app"),
        BooleanField("no_ordering"),
        ListField("syspaths"),
        PluginField("requirements", plugin=RequirementsConfig),
    ]

    @classmethod
    def load(cls, source):
        """
        Loading a manifest source.

        Arguments:
            cls (class): Manifest class.
            source (string or pathlib.Path or dict or Manifest): The Manifest source to
                load. It can be either:

                * A Manifest object, it will just be returned as it without any
                  validation, you are responsible of its correctness;
                * A string for the file path to load in JSON or TOML format;
                * A Path object to the file to load in JSON or TOML format;
                * A Dictionnary which respect the manifest structure;

                Source file format are guessed from their file extension such as JSON
                for ``.json`` or TOML for ``.toml``.

        Return:
            Manifest: A Manifest model instance.
        """
        if isinstance(source, cls):
            # That should commonly not happen but if a Manifest object is given, just
            # use it directly and bypass everything else.
            return source
        elif isinstance(source, dict):
            # A dict is directly used as content
            content = source
        else:
            # Enforce Path object
            source = Path(source)

            # JSON format is the simpliest structure which fit exactly to the manifest
            # structure
            if source.name.endswith(".json"):
                content = json.loads(source.read_text())
            # TOML format is more complex than manifest structure and need to be
            # built correctly
            elif source.name.endswith(".toml"):
                loaded = tomli.loads(source.read_text())

                # Check required structure
                if (
                    "tool" not in loaded or
                    "project_composer" not in loaded["tool"]
                ):
                    raise ComposerManifestError(
                        "TOML manifest must have a section [tool.project_composer] to "
                        "fill base options."
                    )

                content = loaded["tool"]["project_composer"]

                # If there is no composer config name, try to use the TOML project one
                if not content.get("name"):
                    pyproject_name = loaded.get("project", {}).get("name")
                    if pyproject_name:
                        content["name"] = pyproject_name

            # No recognized format
            else:
                raise ComposerManifestError(
                    "Unable to guess the manifest file format. Please suffix your "
                    "filename either with '.json' or '.toml' depending it is a JSON "
                    "or a TOML format."
                )

        # Build Manitest keyword arguments from retrieved content, only retains items
        # knowed as manifest fields
        kwargs = {
            k: v
            for k, v in content.items()
            if k in cls.get_fieldnames()
        }

        return cls(**kwargs)



"""
Specific application exceptions.
"""


class ProjectComposerException(Exception):
    """
    Exception base.

    You should never use it directly except for test purpose. Instead make or
    use a dedicated exception related to the error context.
    """
    pass


class ComposerError(ProjectComposerException):
    """
    Error occuring from a Composer class or inheriter.
    """
    pass


class ComposerConfigError(ProjectComposerException):
    """
    Error occuring from a BaseConfig class or inheriter.
    """
    pass


class ComposerManifestError(ProjectComposerException):
    """
    Error occuring from a manifest.
    """
    pass


class ComposerAppStoreError(ProjectComposerException):
    """
    Error occuring from Application store.
    """
    pass


class ComposerProcessorError(ProjectComposerException):
    """
    Error occuring from a processor.
    """
    pass


class ComposerPurgeError(ComposerProcessorError):
    """
    Error occuring when trying to purge an application repository.
    """
    pass



from ...logger import LoggerBase


class ApplicationUrlCollector(LoggerBase):
    """
    Application urls collector is a class to inherit from an Application url class to
    create url patterns collections.
    """
    def __init__(self, settings=None):
        self.settings = settings

        super().__init__()

    def load_urlpatterns(self, urlpatterns):
        """
        Method to implement by Application Url classes.

        Every classes should not forget to use ``super().load_urlpatterns(urlpatterns)``
        in their ``load_urlpatterns`` method implementation, commonly at the beggining.
        """
        return urlpatterns

    def collect(self, urlpatterns=None):
        self.log.debug("Application urls collector processing")
        urlpatterns = urlpatterns or []
        patterns = self.load_urlpatterns(urlpatterns)

        return patterns



import logging

from . import __pkgname__


def init_logger(name, level, printout=True):
    """
    Initialize app logger to configure its level/handler/formatter/etc..

    Arguments:
        name (str): Logger name used to instanciate and retrieve it.
        level (str): Level name (``debug``, ``info``, etc..) to enable.

    Keyword Arguments:
        printout (bool): If False, logs will never be outputed.

    Returns:
        logging.Logger: Application logger.
    """
    root_logger = logging.getLogger(name)
    root_logger.setLevel(level)

    # Redirect outputs to the void space, mostly for usage within unittests
    if not printout:
        from io import StringIO
        dummystream = StringIO()
        handler = logging.StreamHandler(dummystream)
    # Standard output with optional colored messages
    else:
        try:
            import colorlog
        # If colorlog is not available, turn back to standard formatter
        except ImportError:
            handler = logging.StreamHandler()
            handler.setFormatter(
                logging.Formatter(
                    "%(asctime)s - %(levelname)s - %(message)s",
                    datefmt="%H:%M:%S"
                )
            )
        # If colorlog is available use its formatter
        else:
            handler = logging.StreamHandler()
            handler.setFormatter(
                colorlog.ColoredFormatter(
                    "%(asctime)s - %(log_color)s%(message)s",
                    datefmt="%H:%M:%S"
                )
            )

    root_logger.addHandler(handler)

    return root_logger


class LoggerBase:
    """
    A basic class just to ship the required logger object.

    This class should be at the last position in inheritance definition, since it must
    be called first because next classes may require its ``self.log`` attribute.
    """
    def __init__(self, *args, **kwargs):
        self.log = logging.getLogger(__pkgname__)




class EnabledApplicationMarker:
    """
    Application elligibility marker.

    An application part class must inherit from this to be marked as enabled to load
    for composition.

    This class implements an attribute ``_ENABLED_COMPOSABLE_APPLICATION`` used to
    validate elligibility and nothing else.

    Only purpose of attribute ``_ENABLED_COMPOSABLE_APPLICATION`` is to mark it as
    elligible for composer inspection, don't change it.

    You may however, reproduce elligibility yourself in your class with including this
    attribute, its value is not important except it must not be ``None``.
    """
    _ENABLED_COMPOSABLE_APPLICATION = True



"""
Default values for Diagram args and kwargs
"""
from pathlib import Path

from PIL import Image


STATIC_DIR = Path("../docs/_static/")


def get_conf(filepath, horizontal=True, extra=None):
    """
    Build options for 'diagrams.Diagrams'.

    Default option configuration build a PNG image with transparent background where
    items are aligned from left to the right.

    Arguments:
        filepath (pathlib.Path): Path object for output file destination.

    Keyword Arguments:
        horizontal (boolean): If True, the main direction "left to right" is
            used. If False, the main direction "top to bottom" is used. Default to
            True. Keep in mind that graphviz ordering is special and you may probably
            not retrieve the same item order from horizontal to vertical direction.
        extra (dict): A dictionnary to override items from built options dict.

    Returns:
        dict: Dictionnary of Diagrams options plus additional item 'output_filepath'.
        You will have to pop this additional item before giving options to Diagrams
        since 'output_filepath' is not valid keyword argument.
    """
    conf = {
        "filename": STATIC_DIR / filepath,
        "show": False,
        "direction": "LR" if horizontal else "TB",
        "outformat": "png",
        "graph_attr": {
            "bgcolor": "transparent",
        },
    }

    # If format name is defined, use it to guess output filename
    if conf.get("outformat"):
        conf["output_filepath"] = conf["filename"].with_suffix("." + conf["outformat"])
    # Else just adopt the original filename (wsithout any extension)
    else:
        conf["output_filepath"] = conf["filename"]

    if extra:
        conf.update(extra)

    return conf


def adjust_image(filepath, padding=10):
    """
    Adjust image to add padding and enforce white background.

    This will overwrite the given source image.

    Current code is only compatible with RGBA image and so with a PNG format.

    Arguments:
        filepath (pathlib.Path): Source path object.

    Keyword Arguments:
        padding (integer): Padding size in pixel, applied as both horizontal and
            vertical.

    Returns:
        PIL.Image: Final result image object.
    """
    source = Image.open(filepath)

    # Use getbbox to get the non empty content size so we get ride of the empty margins
    cropped = source.crop(source.getbbox())

    # Compute source cropped size to include padding
    width, height = cropped.size
    new_size = (
        width + (padding * 2),
        height + (padding * 2),
    )

    # Build new image from the cropped one plus additional padding
    padded = Image.new(
        cropped.mode,
        new_size,
        color=(255, 255, 255)
    )
    padded.paste(cropped, (padding, padding))

    # Build white background image
    white_background = Image.new(
        "RGBA",
        padded.size,
        color=(255, 255, 255)
    )
    # Merge
    alpha_composite = Image.alpha_composite(white_background, padded)
    alpha_composite.save(filepath)

    return alpha_composite




class BaseField:
    """
    Field class abstract.

    Arguments:
        name (string): The field name.

    Keyword Arguments:
        default (object): Default value when field is empty.
        required (boolean): True if the field is required to be set.
        plugin (BasePluginConfig): The plugin object where to store plugin fields.

    Attributes:
        _DEFAULT_VALUE (object): Default value to use when no one is explicitely given
            from arguments. It must fit to the field type. By default this is ``None``.
    """
    _DEFAULT_VALUE = None
    TYPE = object

    def __init__(self, name, default=None, required=False, plugin=None):
        self.name = name
        self.default = self.get_default(default)
        self.required = required
        self.plugin = plugin

    def __repr__(self):
        return "<{klass}: {name}>".format(
            klass=self.__class__.__name__,
            name=self.name
        )

    def __str__(self):
        return self.name

    def get_default(self, default=None):
        """
        Return default value.

        Keyword Arguments:
            default (object): Default value according to the field type.

        Returns:
            object: Either default value given as argument if any, else the default
                field type value.
        """
        return default or self._DEFAULT_VALUE

    def fieldtype(self):
        """
        Return the Class name that is used as the field type.
        """
        return self.__class__.__name__


class CharField(BaseField):
    """
    For simple string value.
    """
    _DEFAULT_VALUE = None
    TYPE = str


class ListField(BaseField):
    """
    For a list value.
    """
    _DEFAULT_VALUE = []
    TYPE = list


class BooleanField(BaseField):
    """
    For a boolean value.
    """
    _DEFAULT_VALUE = False
    TYPE = bool


class PluginField(BaseField):
    """
    For a plugin value.

    A plugin is a special field that include sub configuration object.
    """
    _DEFAULT_VALUE = {}
    TYPE = dict

    def __init__(self, *args, plugin, **kwargs):
        kwargs["plugin"] = plugin
        super().__init__(*args, **kwargs)



from __init__ import get_conf, adjust_image
from icons import (
    ApplicationStore, ApplicationRepository, Application,
    Resolver, Processor, ProcessorCaller, Project,
)


if __name__ == "__main__":
    from pathlib import Path

    from diagrams import Cluster, Diagram

    conf = get_conf(Path(__file__).stem)
    output_filepath = conf.pop("output_filepath")

    with Diagram("", **conf):
        repository = ApplicationRepository("Application repository")

        with Cluster("Composer"):
            appstore = ApplicationStore("Application store")
            resolver = Resolver("Resolver")
            proc_caller = ProcessorCaller("Processor caller")

            resolver >> appstore
            appstore >> proc_caller

        repository >> resolver

        with Cluster("Processors"):
            proc_b = Processor("Processor B")
            proc_a = Processor("Processor A")

            processors = [
                proc_a,
                proc_b,
            ]

            proc_caller >> processors

            with Cluster("Part B"):
                part_b_2 = Application("App2 part B")
                part_b_1 = Application("App1 part B")

                part_b = [
                    part_b_2,
                    part_b_1,
                ]

                proc_b >> part_b

            with Cluster("Part A"):
                part_a_2 = Application("App2 part A")
                part_a_1 = Application("App1 part A")

                part_a = [
                    part_a_2,
                    part_a_1,
                ]

                proc_a >> part_a

        project = Project("Project")
        part_a >> project
        part_b >> project

    adjust_image(output_filepath.resolve())
    print("Built diagram at: {}".format(output_filepath.resolve()))



from ...processors import ClassProcessor


class DjangoSettingsProcessor(ClassProcessor):
    """
    Processor for enabled application settings classes for a Django project.
    """
    def get_module_path(self, name):
        """
        Return a Python path for a module name.

        Arguments:
            name (string): Module name.

        Returns:
            string: Module name prefixed with repository path if it is not empty else
            returns just the module name.
        """
        return "{base}.{part}".format(
            base=self.composer.get_application_base_module_path(name),
            part="settings",
        )


class DjangoUrlsProcessor(ClassProcessor):
    """
    Processor for enabled application urls classes for a Django project.
    """
    def get_module_path(self, name):
        """
        Return a Python path for a module name.

        Arguments:
            name (string): Module name.

        Returns:
            string: Module name prefixed with repository path if it is not empty else
            returns just the module name.
        """
        return "{base}.{part}".format(
            base=self.composer.get_application_base_module_path(name),
            part="urls",
        )




def project_settings(composer, base_classes=None, name=None):
    """
    Build composed settings class for given composer.

    Arguments:
        composer (project_composer.compose.Composer): Composer instance.

    Keyword Arguments:
        base_classes (list): A list of base classes inheritage to build the settings
            class. Default to empty list.
        name (string): Class name to set. Default to ``ComposedProjectSettings``.

    Returns:
        object: Composed settings class.
    """
    base_classes = base_classes or []
    name = name or "ComposedProjectSettings"

    # Search for all enabled classes
    classes = composer.call_processor("DjangoSettingsProcessor", "export")

    # Reverse the list since Python class inheritance order start from the last
    # statement to the first
    classes.reverse()

    # Return built settings class from composed settings classes
    return type(name, tuple(classes + base_classes), {})


def project_urls(composer, settings, base_classes=None, name=None):
    """
    Build composed urls collector class for given composer.

    Arguments:
        composer (project_composer.compose.Composer): Composer instance.
        settings (django.conf.settings): Django settings to give to the collector.

    Keyword Arguments:
        base_classes (list): A list of base classes inheritage to build the settings
            class. Default to empty list.
        name (string): Class name to set. Default to ``ComposedProjectSettings``.

    Returns:
        list: List of collected url patterns (like ``django.urls.path`` or
        ``django.urls.re_path``) from all application urls classes.
    """
    base_classes = base_classes or []
    name = name or "ComposedProjectUrls"

    # Search for all enabled classes
    classes = composer.call_processor("DjangoUrlsProcessor", "export")

    # Build Urls class from composed urls
    composed = type(name, tuple(classes + base_classes), {})

    # Collect and return applications urls
    mounter = composed(settings)
    return mounter.collect()



from .base import BasePluginConfig
from .fields import CharField


class RequirementsConfig(BasePluginConfig):
    """
    Requirements file plugin.
    """
    _DEFAULT_INTRO = (
        "# This file is automatically overwritten by composer, DO NOT EDIT IT.\n"
        "# Written on: {creation_date}\n\n"
    )
    _DEFAULT_CONTENT_FILENAME = "requirements.txt"
    _FIELDS = [
        CharField("application_label"),
        CharField("application_divider"),
        CharField("introduction", default=_DEFAULT_INTRO),
        CharField("source_filename", default=_DEFAULT_CONTENT_FILENAME),
        CharField("template"),
    ]



import inspect

from ..exceptions import ComposerConfigError
from .fields import PluginField


class BaseConfig:
    """
    Configuration class abstract.

    Attributes:
        _FIELDS (list): Required list of enabled configuration fields.
    """
    _FIELDS = []

    def __init__(self, *args, **kwargs):
        self.validate_attributes(**kwargs)
        self.install_attributes(**kwargs)

    @classmethod
    def validate_attributes(cls, **kwargs):
        """
        Validate given data against field descriptions.

        Although it's a classmethod, it is safe to call it as an instance method.

        Arguments:
            **kwargs: Fields datas.
        """
        if inspect.isclass(cls):
            klassname = cls.__name__
        else:
            klassname = cls.__class__.__name__

        for field in cls.get_fields():
            # Validate required field
            if field.required and field.name not in kwargs:
                msg = "Field '{field}' is required from '{klass}'"
                raise ComposerConfigError(msg.format(
                    field=field.name,
                    klass=klassname,
                ))

            # Validate given value against field type
            if (
                kwargs.get(field.name) is not None and
                not isinstance(kwargs.get(field.name), field.TYPE) and
                not isinstance(kwargs.get(field.name), BasePluginConfig)
            ):
                msg = "'{klass}' field '{field}' must be a '{type}' not '{wrong}'"
                raise ComposerConfigError(msg.format(
                    field=field.name,
                    type=field.TYPE.__name__,
                    klass=klassname,
                    wrong=type(kwargs.get(field.name)).__name__,
                ))

    def install_attributes(self, **kwargs):
        """
        Install defined field as object attributes with given possible values from
        kwargs.

        ``BaseConfig.validate_attributes()`` must be runned before to ensure field
        values are valid.

        Arguments:
            **kwargs: Keyword arguments for field values to set as object attribute
                value.
        """
        for field in self.get_fields():
            value = field.default
            if kwargs.get(field.name) is not None:
                value = kwargs.get(field.name)

            # Field plugin needs to build its plugin object before setting its
            # attributes
            if isinstance(field, PluginField):
                if isinstance(value, BaseConfig):
                    plugin = value
                else:
                    plugin = field.plugin()
                    plugin.install_attributes(**value)
                setattr(self, field.name, plugin)
            else:
                setattr(self, field.name, value)

    @classmethod
    def get_fields(cls):
        """
        Return model field definitions.

        Arguments:
            cls (class or object): Config class or object.

        Returns:
            list: Lists every defined field object.
        """
        return cls._FIELDS

    @classmethod
    def get_fieldnames(cls):
        """
        Return model field names.

        Arguments:
            cls (class or object): Config class or object.

        Returns:
            list: Lists every defined field name.
        """
        return [item.name for item in cls.get_fields()]

    @classmethod
    def get_fieldtypes(cls):
        """
        Return model field types.

        Arguments:
            cls (class or object): Config class or object.

        Returns:
            list: List every defined field type name.
        """
        return [item.fieldtype() for item in cls.get_fields()]

    def to_dict(self):
        """
        Dump manifest values as Python dictionnary.

        Recursively walk in ``to_dict`` method of item that are a children of
        BaseConfig.

        Returns:
            dict: Dictionnary of all field values, including the "requirements"
            ones.
        """
        content = {}

        for name in self.get_fieldnames():
            attr = getattr(self, name)
            if isinstance(attr, BasePluginConfig):
                content[name] = attr.to_dict()
            else:
                content[name] = getattr(self, name)

        return content


class BasePluginConfig(BaseConfig):
    """
    Plugin configuration abstract.
    """
    pass



from pathlib import Path

import click

from ..defaults import DEFAULT_MANIFEST_PATH


# Shared options
COMMON_OPTIONS = {
    "manifest": {
        "args": ("--manifest",),
        "kwargs": {
            "type": click.Path(
                exists=True,
                file_okay=True,
                dir_okay=False,
                path_type=Path
            ),
            "default": DEFAULT_MANIFEST_PATH,
            "metavar": "FILEPATH",
            "help": (
                "Required file path to a project manifest."
            ),
        }
    },
    "repository": {
        "args": ("--repository",),
        "kwargs": {
            "default": None,
            "metavar": "PYTHONPATH",
            "help": (
                "A Python path (aka: 'foo.apps') where to search for application "
                "modules. Using this argument will override the 'repository' item from"
                "manifest."
            ),
        }
    },
    "syspaths": {
        "args": ("--syspath", "syspaths"),
        "kwargs": {
            "default": [],
            "multiple": True,
            "metavar": "DIRPATH",
            "help": (
                "Path to a directory to add to 'sys.path' and which contains Python "
                "modules required by some applications if not already available in the "
                "scope of this tool. You can define it multiple times for each needed "
                "path. Using this argument will override the 'syspaths' value from "
                "manifest."
            ),
        }
    },
}



from __init__ import get_conf, adjust_image
from icons import (
    AppNode, Application, Manifest, Resolver,
    Processor, ProcessorCaller, Project, ProjectPart,
)


if __name__ == "__main__":
    from pathlib import Path

    from diagrams import Cluster, Diagram

    conf = get_conf(Path(__file__).stem)
    output_filepath = conf.pop("output_filepath")

    with Diagram("", **conf):
        with Cluster("Application repository"):
            app_1 = Application("Application 1")
            app_3 = Application("Application 3")
            app_2 = Application("Application 2")

            repository = [
                app_1,
                app_3,
                app_2,
            ]

        manifest = Manifest("Manifest")
        repository >> manifest

        with Cluster("Application store"):
            appnode_3 = AppNode("AppNode 3")
            appnode_2 = AppNode("AppNode 2")
            appnode_1 = AppNode("AppNode 1")

            appstore = [
                appnode_1,
                appnode_2,
                appnode_3,
            ]

        resolver = Resolver("Resolver")
        manifest >> resolver
        resolver >> appstore

        proc_caller = ProcessorCaller("Processor caller")

        with Cluster("Processors"):
            proc_b = Processor("Processor B")
            proc_a = Processor("Processor A")

            processors = [
                proc_b,
                proc_a,
            ]

            appstore >> proc_caller
            proc_caller >> processors

            with Cluster("Part A"):
                part_a_2 = ProjectPart("App2 part A")
                part_a_1 = ProjectPart("App1 part A")

                part_a = [
                    part_a_2,
                    part_a_1,
                ]

                proc_a >> part_a

            with Cluster("Part B"):
                part_b_2 = ProjectPart("App2 part B")
                part_b_1 = ProjectPart("App1 part B")

                part_b = [
                    part_b_2,
                    part_b_1,
                ]

                proc_b >> part_b

        project = Project("Project")
        part_b >> project
        part_a >> project

    adjust_image(output_filepath.resolve())
    print("Built diagram at: {}".format(output_filepath.resolve()))



import datetime

from pathlib import Path

from .base import ComposerProcessor
from ..exceptions import ComposerProcessorError


class TextContentProcessor(ComposerProcessor):
    """
    Text content composer assemble all text content files from enabled Applications.

    Although it has been done as a generic solution for any content files, this is
    currently tied to specific ``requirements`` plugin from manifest.
    """
    def get_template(self, template=None):
        """
        Get the base content text used to build final content.

        Keyword Arguments:
            template (string or pathlib.Path): Path to file of base content to start
                output. By default there is none.

        Returns:
            string: Base content. If no template has been given, an empty string is
            returned instead.
        """
        if template:
            template = Path(template)
            return template.read_text()

        return ""

    def export(self):
        """
        Combinate all application content files into a single content string.

        Returns:
            string: Combinated content files.
        """
        output = ""

        requirements_config = self.composer.manifest.requirements

        if requirements_config.introduction:
            output += requirements_config.introduction.format(
                creation_date=datetime.datetime.now().isoformat(timespec="seconds"),
            )

        output += self.get_template(requirements_config.template)

        for node in self.composer.apps:
            # Try to find application module
            module_path = self.composer.get_module_path(node.name)
            module = self.composer.find_app_module(module_path)

            if module and getattr(module, "__file__", None):
                # Resolve expected text content file path inside module
                source_path = (
                    Path(module.__file__).parents[0].resolve() /
                    requirements_config.source_filename
                )
                # Try to find file from application to append its content to the output
                if source_path.exists():
                    msg = "{klass} found content file at: {path}".format(
                        klass=self.__class__.__name__,
                        path=source_path,
                    )
                    self.composer.log.debug(msg)

                    content = source_path.read_text()
                    if content.strip():
                        if requirements_config.application_divider:
                            output += requirements_config.application_divider

                        if requirements_config.application_label:
                            label = requirements_config.application_label
                            output += label.format(name=node.name)

                        output += content

                else:
                    msg = "{klass} is unable to find content file from: {path}".format(
                        klass=self.__class__.__name__,
                        path=source_path,
                    )
                    self.composer.log.debug(msg)

        return output

    def dump(self, **kwargs):
        """
        Write export payload to a dump file.

        Arguments:
            destination (pathlib.Path): Path object for the dump file destination.

        Returns:
            pathlib.Path: The Path object where the file has been writed.
        """
        destination = kwargs.get("destination")
        if not destination:
            raise ComposerProcessorError("Keyword argument 'destination' is required")

        output = self.export()
        destination.write_text(output)

        return destination

    def check(self, printer=print):
        """
        Debugging check what this processor should find or match.
        """
        printer()
        printer("🧵 Processor '{}'".format(self.__class__.__name__))

        app_last = len(self.composer.apps)
        for i, node in enumerate(self.composer.apps, start=1):
            # Try to find application module
            module_path = self.composer.get_module_path(node.name)
            module = self.composer.find_app_module(module_path)

            # Display app label name
            printer(
                "X" if (i == app_last) else "T",
                node.name,
            )

            # Try to find app module file to get its path
            if module and getattr(module, "__file__", None):
                # Resolve expected text content file path inside module
                source_path = (
                    Path(module.__file__).parents[0].resolve() /
                    self.composer.manifest.requirements.source_filename
                )

                # Try to find a requirement file
                if source_path.exists():
                    content = source_path.read_text()
                    # Lists package, omits possible commentaries
                    pkgs = [
                        pkg
                        for pkg in content.strip().splitlines()
                        if not pkg.startswith("#")
                    ]
                    pkg_last = len(pkgs)
                    # Display found package lines
                    for p, pkg in enumerate(pkgs, start=1):
                        printer(
                            (
                                "O" if (i == app_last) else "I"
                            ) + (
                                "X" if (p == pkg_last) else "T"
                            ),
                            pkg,
                        )
                # There was no requirement file
                else:
                    printer(
                        "OX" if (i == app_last) else "IX",
                        "No requirement file",
                        yes_or_no=False,
                    )

        return



import click

from .. import __pkgname__, __version__


@click.command()
@click.pass_context
def version_command(context):
    """
    Print out version information.
    """
    click.echo("{name} {version}".format(name=__pkgname__, version=__version__))



import sys
import inspect
from pathlib import Path

from .app_storage import AppStore
from .exceptions import ComposerError
from .importer import import_module
from .logger import LoggerBase
from .manifest import Manifest
from .utils.tree_printer import TreePrinter


class Composer(LoggerBase):
    """
    Composer base implements everything about application module discovering and
    manifest loading.

    Arguments:
        manifest (string or pathlib.Path or dict or Manifest): The Manifest source to
            load. It can be either:

            * A Manifest object, it will just be returned as it without any validation.
              You are responsible of its correctness;
            * A string for the file path to load in JSON or TOML format;
            * A Path object to the file to load in JSON or TOML format;
            * A Dictionnary which respect the manifest structure;

            Source file format are guessed from their file extension such as JSON
            for ``.json`` or TOML for ``.toml``.
        processors (list): List of available composition processors classes.

    Attributes:
        _APPLICATION_MODULE_PYTHONPATH (string): A template string to build the full
            Python path of founded class. It expected two variables ``parent`` and
            ``name``, respectively the module path and the class name.
    """
    _APPLICATION_MODULE_PYTHONPATH = "{parent}.{name}"

    def __init__(self, manifest, processors=[]):
        super().__init__()

        self.manifest = self.get_manifest(manifest)
        self.set_syspaths(self.manifest.syspaths or [])

        self.store = AppStore(default_app=self.manifest.default_store_app)

        self.apps = []

        # Register and initialize all given processors
        self.processors = {
            proc.__name__: proc(self)
            for proc in processors
        }

    def get_manifest(self, manifest):
        """
        Return loaded manifest object.

        Arguments:
            manifest (string or pathlib.Path or dict or Manifest): The Manifest source
                to load.

        Returns:
            Manifest: The manifest object loaded from given source.
        """
        return Manifest.load(manifest)

    def set_syspaths(self, paths):
        """
        Append each item path to ``sys.path``.

        This won't never append a same path twice.

        Arguments:
            paths (list): A list of path to append.
        """
        for path in paths:
            if path not in sys.path:
                sys.path.append(path)

    def get_application_base_module_path(self, name):
        """
        Return the Python path to the application base module.

        Commonly this should be the ``__init__.py`` file from application directory.

        Arguments:
            name (string): Module name.

        Returns:
            string: Module name prefixed with repository path if it is not empty else
            returns just the module name.
        """
        if self.manifest.repository:
            return self._APPLICATION_MODULE_PYTHONPATH.format(
                parent=self.manifest.repository,
                name=name,
            )

        return name

    def get_module_path(self, name):
        """
        Return a Python path for a module name.

        Arguments:
            name (string): Module name.

        Returns:
            string: Module name prefixed with repository path if it is not empty else
            returns just the module name.
        """
        return self.get_application_base_module_path(name)

    def find_app_module(self, name):
        """
        Find a module (by its pythonpath) from application.

        Arguments:
            name (string): Module pythonpath.

        Returns:
            object: Module object if found else None.
        """
        try:
            module = import_module(name)
        except ModuleNotFoundError:
            msg = "{klass} is unable to find module: {path}".format(
                klass=self.__class__.__name__,
                path=name,
            )
            self.log.debug(msg)
            return None
        else:
            return module

    def _is_elligible_class(self, obj):
        """
        Find if given object is an enabled class for composition.

        Criterias for eligibility are in order:

        * Object is a class;
        * Class is not named ``EnabledApplicationMarker``;
        * Class got attribute ``_ENABLED_COMPOSABLE_APPLICATION`` which value is not
          ``None``;

        Arguments:
            obj (object): Object to check for eligibility.

        Returns:
            boolean: ``True`` if object is eligibile to criterias else ``False``.
        """
        if (
            inspect.isclass(obj) and
            getattr(obj, "__name__", None) != "EnabledApplicationMarker" and
            getattr(obj, "_ENABLED_COMPOSABLE_APPLICATION", None) is not None
        ):
            return True

        return False

    def _get_elligible_module_classes(self, path, module):
        """
        Get all elligible classes from a module.

        Arguments:
            path (string): The Python path to a module used for reporting and logging
                messages.
            module (object): The module object where to find elligible classes.

        Returns:
            list: List of elligible classes objects.
        """
        enabled = []

        if not hasattr(module, "__dict__"):
            msg = "Module object from '{}' must have a '__dict__' attribute."
            raise ComposerError(msg.format(path))

        for object_name in module.__dict__.keys():
            if not object_name.startswith("_"):
                obj = getattr(module, object_name)

                if self._is_elligible_class(obj):
                    msg = "{klass} found enabled Class at: {path}.{object_name}".format(
                        klass=self.__class__.__name__,
                        path=path,
                        object_name=object_name,
                    )
                    self.log.debug(msg)

                    if obj not in enabled:
                        enabled.append(obj)

        return enabled

    def _scan_app_module(self, name):
        """
        Load an application module to get its options.

        Arguments:
            path (string): The Python path to a module to check for.

        Returns:
            dict: Application payload (name, dependencies and push_end options).
        """
        path = self.get_module_path(name)

        # Try to find application module and get its possible parameter variables
        module = self.find_app_module(path)
        if module:
            msg = "{klass} found application at: {path}".format(
                klass=self.__class__.__name__,
                path=path,
            )
            self.log.debug(msg)

            payload = {
                "name": name,
                "filepath": module.__file__
            }

            if hasattr(module, "DEPENDENCIES"):
                payload["dependencies"] = getattr(module, "DEPENDENCIES")

            if hasattr(module, "PUSH_END"):
                payload["push_end"] = getattr(module, "PUSH_END")

            return payload

        return None

    def call_processor(self, name, method, **kwargs):
        """
        Execute a processor method.

        Arguments:
            name (string): Processor name in registry.
            method (string): Processor method to execute.
            **kwargs: Keyword arguments to pass to method if any.

        Returns:
            object: Content type depend from what processor method returns.
        """
        if name not in self.processors:
            msg = "Given processor name is not registered from composer: {}"
            raise ComposerError(msg.format(name))

        if not hasattr(self.processors[name], method):
            msg = "Processor '{proc}' don't have any method named '{method}'"
            raise ComposerError(msg.format(proc=name, method=method))

        return getattr(self.processors[name], method)(**kwargs)

    def resolve_collection(self, lazy=True):
        """
        Resolve collection with AppStore.

        Keyword Arguments:
            lazy (boolean): If True, there won't be any dependency order resolving and
                the application list will just be the collection with ``AppNode``
                objects instead of name strings. If False, the resolving will be
                processed. Default is ``True``.

        Returns:
            list: List of ``AppNode`` objects.
        """
        collection = []

        for name in self.manifest.collection:
            payload = self._scan_app_module(name)
            # Ignore unfound application
            if payload:
                collection.append(payload)

        if lazy:
            self.apps = self.store.resolve(
                collection,
                no_ordering=True
            )
        else:
            self.apps = self.store.resolve(
                collection,
                no_ordering=self.manifest.no_ordering
            )

        return collection

    def check(self, lazy=True, printer=None):
        """
        Output some informations about given manifest, app resolving and processors.

        It is strongly recommended that every checking is directly outputted. It means
        you should not build a list of messages to output at the end, instead each job
        should directly output what it has checked.

        This is to ensure the debugging won't hide what have been done before a
        critical error. Remember this method should be almost safe since it is for
        debugging.

        Keyword Arguments:
            lazy (boolean): Wheither to use the lazy mode or not with composer
                resolver.
            printer (callable): A callable to use to output debugging informations.
                Default to builtin function ``utils.tree_printer.TreePrinter`` to
                benefit from the tree alike display.
        """
        printer = printer or TreePrinter(printable=True)

        printer("👷 Checking composer")
        printer()

        printer("📄 Manifest")
        printer("T", "Name: {}".format(self.manifest.name))
        printer("T", "Repository: {}".format(self.manifest.repository))
        printer("X", "Collection:")
        if self.manifest.collection:
            last = len(self.manifest.collection)
            for i, item in enumerate(self.manifest.collection, start=1):
                printer(
                    "OX" if (i == last) else "OT",
                    item,
                )

        # Scan repository
        printer()
        printer("🌐 Repository directory")
        repository_mod = self.find_app_module(self.manifest.repository)
        if repository_mod:
            repository_dirpath = Path(repository_mod.__file__).parents[0]

            printer("X", repository_dirpath, yes_or_no=repository_dirpath.exists())

            last = len(self.manifest.collection)
            for i, item in enumerate(self.manifest.collection, start=1):
                app_dirpath = repository_dirpath / item
                printer(
                    "OX" if (i == last) else "OT",
                    app_dirpath,
                    yes_or_no=app_dirpath.exists(),
                )
        else:
            printer("X", "Unable to find repository directory")

        self.resolve_collection(lazy=lazy)

        printer()
        printer("🗃️ Resolved applications")
        if self.apps:
            last = len(self.apps)
            for i, app in enumerate(self.apps, start=1):
                printer(
                    "X" if (i == last) else "T",
                    app,
                )
                # Push end option
                printer(
                    "OT" if (i == last) else "IT",
                    "Push end: {}".format(app.push_end),
                )
                # Dependencies
                dep_msg = "No dependency"
                if app.dependency_names:
                    dep_msg = "Dependencies: {}".format(", ".join(app.dependency_names))
                printer(
                    "OX" if (i == last) else "IX",
                    dep_msg,
                )
        else:
            printer("X", "Unable to find any applications")

        # Call for processors check
        for name, proc in self.processors.items():
            proc.check(printer=printer)



"""
This contains dummy proxy objects to the right icons since it currently use GCP icons
but is awaiting for a better and more accurate icon set.
"""
from diagrams.gcp.analytics import Composer as GCPComposer, Pubsub
from diagrams.gcp.compute import GKEOnPrem, ComputeEngine, GPU
from diagrams.gcp.database import Memorystore
from diagrams.gcp.devtools import ContainerRegistry, IdePlugins, SDK
from diagrams.gcp.ml import AdvancedSolutionsLab, AIHub
from diagrams.gcp.storage import Filestore


class Manifest(Filestore):
    pass


class ApplicationStore(AIHub):
    pass


class ApplicationRepository(ContainerRegistry):
    pass


class AppNode(SDK):
    pass


class Application(IdePlugins):
    pass


class Composer(GCPComposer):
    pass


class Resolver(Pubsub):
    pass


class ProcessorCaller(Memorystore):
    pass


class Processor(ComputeEngine):
    pass


class ProjectPart(GPU):
    pass


class PackageInstaller(GKEOnPrem):
    pass


class Project(AdvancedSolutionsLab):
    pass


if __name__ == "__main__":
    from pathlib import Path

    from diagrams import Cluster, Diagram

    from __init__ import get_conf, adjust_image

    conf = get_conf(Path(__file__).stem)
    output_filepath = conf.pop("output_filepath")

    with Diagram("", **conf):
        with Cluster("GCP set"):
            with Cluster("Sources of truth"):
                manifest = Manifest("Manifest")

            with Cluster("Applications"):
                appstore = ApplicationStore("Application store")
                repository = ApplicationRepository("Application repository")
                appnode = AppNode("AppNode")
                application = Application("Application")

            with Cluster("Composer mechanics"):
                composer = Composer("Composer")
                resolver = Resolver("Resolver")
                proc_caller = ProcessorCaller("Processor caller")

            with Cluster("Processors"):
                processor = Processor("Processor")
                project_part = ProjectPart("Project part")

            with Cluster("Consumers"):
                packager_installer = PackageInstaller("Package installer")
                project = Project("Project")

    adjust_image(output_filepath.resolve())
    print("Built diagram at: {}".format(output_filepath.resolve()))




class TreePrinter:
    """
    Shortcut to make tree row output.

    This is useful to display a tree without a real tree implementation.

    It allows to display tree rows each one after one. However this technique is not
    very smart and so you need to give it a pattern that it will correctly pad and
    translate to the right glyphs.

    Usage example: ::

        treeprint = TreePrinter()

        treeprint("Plop")
        >>>Plop
        treeprint("T", "Plap")
        >>>├──Plap
        treeprint("T", "Plip")
        >>>├──Plip
        treeprint("IT", "Ping")
        >>>│  ├──Ping
        treeprint("IIX", "Zip")
        >>>│  │  └──Zip
        treeprint("IIOT", "Zap")
        >>>│  │     └──Zap
        treeprint("IIOIT", "Yyyy")
        >>>│  │        ├──Yyyy
        treeprint("IIOIX", "Zzzz")
        >>>│  │        └──Zzzz
        treeprint("IIOX", "Zup")
        >>>│  │     └──Zup
        treeprint("IX", "Pong")
        >>>│  └──Pong
        treeprint("X", "Plop")
        >>>└──Plop

    """
    TEE_CHAR = "├"
    ELBOW_CHAR = "└"
    LINE_CHAR = "── "
    YESLINE_CHAR = "─✔ "
    NOLINE_CHAR = "─✖ "
    PIPE_CHAR = "│"
    VOID_CHAR = " "

    def __init__(self, printable=False):
        self.store = {}
        self.current_level = 0
        self.spectrum = ""
        self.printable = printable

    def __call__(self, *args, **kwargs):
        if self.printable:
            print(self.build(*args, **kwargs))
        else:
            return self.build(*args, **kwargs)

    def build(self, *args, **kwargs):
        """
        Build the ascii row for given content.

        Arguments:
            *args: Either just the content without pattern. Or the pattern then the
                content.

        Returns:
            string: The ascii row glyphs followed by given content.
        """
        if len(args) == 0:
            return ""
        elif len(args) > 1:
            pattern = args[0]
            content = args[1]
        else:
            pattern = ""
            content = args[0]

        yes_or_no = kwargs.get("yes_or_no", None)

        # Enforce string
        content = str(content)

        return self.get_indent(pattern, yes_or_no=yes_or_no) + content

    def get_indent(self, pattern, yes_or_no=None):
        """
        Convert a pattern to an ascii tree row

        * "T" is for an intersection part, when the current item have sibling;
        * "X" is for an end part, when current item have no sibling;
        * "O" is for a void part, when there is parent level without sibling;
        * "I" is for a pipe part, when there is parent with sibling;

        Arguments:
            pattern (string): A pattern composed of letters used to determined the
                ascii representation to append to output.

        Keyword Arguments:
            yes_or_no (boolean): If True use YESLINE_CHAR, if false use NOLINE_CHAR and
                finally if None (default) use LINE_CHAR for intersection and end parts.

        Returns:
            string: Ascii representation of pattern.
        """
        output = ""

        line_char = self.LINE_CHAR
        if yes_or_no is True:
            line_char = self.YESLINE_CHAR
        elif yes_or_no is False:
            line_char = self.NOLINE_CHAR

        if pattern:
            for item in pattern.upper():
                if item == "T":
                    output += self.TEE_CHAR + line_char
                elif item == "I":
                    output += self.PIPE_CHAR + (" " * len(self.LINE_CHAR))
                elif item == "O":
                    output += self.VOID_CHAR + (" " * len(self.LINE_CHAR))
                elif item == "X":
                    output += self.ELBOW_CHAR + line_char

        return output



import logging

import click

from .. import __pkgname__

from ..compose import Composer
from ..manifest import Manifest
from ..processors import PurgeProcessor

from .base_options import COMMON_OPTIONS


@click.command()
@click.option(
    *COMMON_OPTIONS["manifest"]["args"],
    **COMMON_OPTIONS["manifest"]["kwargs"]
)
@click.option(
    *COMMON_OPTIONS["repository"]["args"],
    **COMMON_OPTIONS["repository"]["kwargs"]
)
@click.option(
    *COMMON_OPTIONS["syspaths"]["args"],
    **COMMON_OPTIONS["syspaths"]["kwargs"]
)
@click.option(
    "--commit",
    is_flag=True,
    help=(
        "Commit the purge."
    ),
)
@click.pass_context
def purge_command(*args, **parameters):
    """
    Purge applications repository from any directory that is not an enabled application
    module.

    Also the command cannot purge application directories that are not in application
    repository.

    WARNING: Once removed, the application directories cannot be retrieved.
    """
    logger = logging.getLogger(__pkgname__)

    # Load manifest settings
    logger.debug("Using manifest: {}".format(parameters["manifest"]))
    manifest = Manifest.load(parameters["manifest"])

    # Patch arguments in multiple mode since an empty default list trouble the
    # manifest settings overriding
    if len(parameters.get("syspaths", [])) == 0:
        parameters["syspaths"] = None

    # Override base manifest settings from given arguments
    # syspaths management have a special thing to avoid default value (empty list) to
    # override the manifest value
    for name in manifest.get_fieldnames():
        if (name != "requirements" and parameters.get(name) is not None):
            setattr(manifest, name, parameters.get(name))

    if not manifest.repository:
        logger.critical(
            "Applications repository is required for this command, read help for "
            "more details."
        )
        raise click.Abort()

    # Logging used settings
    logger.debug("Applications repository: {}".format(manifest.repository))

    for item in manifest.syspaths:
        logger.debug("Loading in sys.path: {}".format(item))

    composer = Composer(manifest, processors=[PurgeProcessor])
    composer.resolve_collection(lazy=False)

    commit = parameters.get("commit")
    if commit:
        to_remove = composer.call_processor("PurgeProcessor", "commit")
    else:
        to_remove = composer.call_processor("PurgeProcessor", "export")
        for path in to_remove:
            msg = "This application module would be removed: {}"
            logger.info(msg.format(path))

    if not to_remove:
        logger.warning("There was not any application module to remove")



from .base import ComposerProcessor


class ClassProcessor(ComposerProcessor):
    """
    Class composer find all existing classes for enabled application modules and that
    match criterias from ``Composer._is_elligible_class``.
    """
    def export(self, **kwargs):
        """
        Export enabled applications classes.

        Returns:
            list: A list of Python classes found as elligible for criterias. The list
            is firstly ordered by the order of enabled application from manifest and
            secondly by their definition order in their module (if there is two classes
            defined with the same name, the first is retained and the second one is
            ignored).
        """
        mods = []
        _mod_names = set([])

        for node in self.composer.apps:
            path = self.get_module_path(node.name)

            # Try to find module
            module = self.composer.find_app_module(path)
            if module:
                msg = "{klass} found module at: {path}".format(
                    klass=self.__class__.__name__,
                    path=path,
                )
                self.composer.log.debug(msg)

                mods.extend([
                    item
                    for item in self.composer._get_elligible_module_classes(path,
                                                                            module)
                    if item.__name__ not in _mod_names
                ])
                # Update the list of unique retained class names which are used
                # for previous uniqueness comparaison in next iteration
                _mod_names.update([item.__name__ for item in mods])

        return mods

    def check(self, printer=print):
        """
        Debugging check what this processor should find or match.

        Keyword Arguments:
            printer (callable): A callable to use to output debugging informations.
                Default to builtin function ``print`` but it won't be very pretty,
                we recommend to use ``utils.tree_printer.TreePrinter`` to benefit from
                the tree alike display. Note than composer already give ``TreePrinter``
                to this argument when calling this method.
        """
        printer()
        printer("🧵 Processor '{}'".format(self.__class__.__name__))

        app_last = len(self.composer.apps)
        for i, node in enumerate(self.composer.apps, start=1):
            printer(
                "X" if (i == app_last) else "T",
                node.name
            )
            path = self.get_module_path(node.name)
            module = self.composer.find_app_module(path)

            if module:
                klasses = self.composer._get_elligible_module_classes(path, module)
                klass_last = len(klasses)
                for k, item in enumerate(klasses, start=1):
                    printer(
                        (
                            "O" if (i == app_last) else "I"
                        ) + (
                            "X" if (k == klass_last) else "T"
                        ),
                        item.__name__
                    )
            else:
                pass



"""
Convenient way to programmatically import a module from a Python path.

"""
import importlib
import sys


def import_module(name, package=None):
    """
    An approximate implementation of import taken from Python importlib documentation.

    This will import a module from given python path in name argument. package argument
    can be given to search in. If module is not in current directory, you will have to
    add its parent path in sys.path. Also relative module path will work only for
    inside a package.

    Arguments:
        name (string): A valid Python path to the module.

    Keyword Arguments:
        package (string): A package name to exclusively search in package tree.

    Returns
        object: Found module object.
    """
    absolute_name = importlib.util.resolve_name(name, package)

    try:
        return sys.modules[absolute_name]
    except KeyError:
        pass

    path = None
    if '.' in absolute_name:
        parent_name, _, child_name = absolute_name.rpartition('.')
        parent_module = import_module(parent_name)
        path = parent_module.__spec__.submodule_search_locations

    for finder in sys.meta_path:
        # Old Meta path finders made for "imp" did not implement the "find_spec" as
        # required with importlib
        if hasattr(finder, "find_spec"):
            spec = finder.find_spec(absolute_name, path)
            # Found module from a loader use it and stop to search
            if spec is not None:
                break
        else:
            continue
    else:
        msg = f'No module named {absolute_name!r}'
        raise ModuleNotFoundError(msg, name=absolute_name)

    module = importlib.util.module_from_spec(spec)
    sys.modules[absolute_name] = module
    spec.loader.exec_module(module)

    if path is not None:
        setattr(parent_module, child_name, module)

    return module



from pathlib import Path

from project_composer.compose import Composer


def check_project(manifest, processors=[], lazy=True):
    """
    Launch composer debugging on a project.

    Composer and processor check method do not return anything and just print out
    every debugging output.

    Arguments:
        manifest (pathlib.Path): Manifest Path object.

    Keyword Arguments:
        processors (list): List of processor to enable in composer.
        lazy (boolean): The composer ``lazy`` argument value to use.

    """
    manifest = Path(manifest).resolve()

    _composer = Composer(
        manifest,
        processors=processors,
    )

    _composer.check(lazy=lazy)



from __init__ import get_conf, adjust_image
from icons import AppNode, Application, Manifest, Resolver


if __name__ == "__main__":
    from pathlib import Path

    from diagrams import Cluster, Diagram

    conf = get_conf(Path(__file__).stem)
    output_filepath = conf.pop("output_filepath")

    with Diagram("", **conf) as diag:
        manifest = Manifest("Manifest")
        resolver = Resolver("Resolver")

        with Cluster("Application repository"):
            app_1 = Application("Application 1")
            app_3 = Application("Application 3")
            app_2 = Application("Application 2")

            repository = [
                app_1,
                app_3,
                app_2,
            ]

        repository >> manifest

        with Cluster("Application store"):
            appnode_1 = AppNode("AppNode 1")
            appnode_2 = AppNode("AppNode 2")
            appnode_3 = AppNode("AppNode 3")

            appstore = [
                appnode_1,
                appnode_2,
                appnode_3,
            ]

        manifest >> resolver
        resolver >> appstore

    adjust_image(output_filepath.resolve())
    print("Built diagram at: {}".format(output_filepath.resolve()))



import logging
from pathlib import Path

import click

from .. import __pkgname__

from ..compose import Composer
from ..manifest import Manifest
from ..processors import TextContentProcessor

from .base_options import COMMON_OPTIONS


@click.command()
@click.option(
    *COMMON_OPTIONS["manifest"]["args"],
    **COMMON_OPTIONS["manifest"]["kwargs"]
)
@click.option(
    *COMMON_OPTIONS["repository"]["args"],
    **COMMON_OPTIONS["repository"]["kwargs"]
)
@click.option(
    *COMMON_OPTIONS["syspaths"]["args"],
    **COMMON_OPTIONS["syspaths"]["kwargs"]
)
@click.option(
    "--template",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=Path),
    default=None,
    metavar="FILEPATH",
    help=(
        "File path to a base requirements file which will be extended (not "
        "overwritten) with application requirements. Ignore this argument if you "
        "want to start from scratch."
    ),
)
@click.option(
    "--dump",
    type=click.Path(exists=False, file_okay=True, dir_okay=False, path_type=Path),
    default=None,
    metavar="FILEPATH",
    help=(
        "File path where to dump exported requirements combination from applications."
    ),
)
@click.option(
    "--applabel",
    "application_label",
    default=None,
    metavar="STRING",
    help=(
        "A string to add the application name before its requirements. The name will "
        "be added in expected pattern '{name}' Remember to finish you string with a"
        "newline character.."
    ),
)
@click.option(
    "--appdivider",
    "application_divider",
    default=None,
    metavar="STRING",
    help=(
        "String to add between application. Commonly used to insert a new line. "
        "By default there is no divider."
    ),
)
@click.option(
    "--introduction",
    default=None,
    metavar="STRING",
    help=(
        "Introduction string to start output. Accept a pattern '{creation_date}' to "
        "include the current date. Give an empty string to disable introduction. If "
        "argument is not defined a default introduction is included with a WARNING "
        "and the date."
    ),
)
@click.option(
    "--source",
    "source_filename",
    default=None,
    metavar="STRING",
    help=(
        "Filename to search in application module to find requirements."
    ),
)
@click.pass_context
def requirements_command(*args, **parameters):
    """
    Output composed requirements from applications.
    """
    logger = logging.getLogger(__pkgname__)

    # Load manifest settings
    logger.debug("Using manifest: {}".format(parameters["manifest"]))
    manifest = Manifest.load(parameters["manifest"])

    # Patch arguments in multiple mode since they empty default list trouble the
    # manifest settings overriding
    if len(parameters.get("syspaths", [])) == 0:
        parameters["syspaths"] = None

    # Override base manifest settings from given arguments
    # syspaths management have a special thing to avoid default value (empty list) to
    # override the manifest value
    for name in manifest.get_fieldnames():
        if (name != "requirements" and parameters.get(name) is not None):
            setattr(manifest, name, parameters.get(name))

    # The same for requirements plugin
    for name in manifest.requirements.get_fieldnames():
        if parameters.get(name) is not None:
            setattr(manifest.requirements, name, parameters.get(name))

    # Logging used settings
    if manifest.repository:
        logger.debug("Applications repository: {}".format(manifest.repository))

    for item in manifest.syspaths:
        logger.debug("Loading in sys.path: {}".format(item))

    logger.debug("Using template: {}".format(manifest.requirements.template))
    logger.debug("Using source filename: {}".format(
        manifest.requirements.source_filename)
    )

    composer = Composer(manifest, processors=[TextContentProcessor])
    composer.resolve_collection(lazy=False)

    dump = parameters.get("dump")
    if dump:
        logger.debug("Dump destination: {}".format(dump))
        msg = "Requirements file written at: {}"
        logger.debug(msg.format(
            composer.call_processor("TextContentProcessor", "dump", destination=dump)
        ))
    else:
        click.echo(composer.call_processor("TextContentProcessor", "export"))




class ComposerProcessor:
    """
    The base processor class to implement.

    This processor does not do anything nor have any real methods to call from
    composer.

    Arguments:
        composer (Composer): The composer instance where this processor will be
            executed.

    Attributes:
        composer (Composer): The composer instance used by processor to get manifest
            object and some internal composer methods to work with application
            repository.
    """
    def __init__(self, composer):
        self.composer = composer

    def get_module_path(self, name):
        """
        Return a Python path for a specific module name base on base application path
        from composer.

        This default implementation just use the base application module (``__init__``).
        Other processors may override it to use a specific module.

        Arguments:
            name (string): Module name.

        Returns:
            string: Module name prefixed with repository path if it is not empty else
            returns just the module name.
        """
        return self.composer.get_application_base_module_path(name)

    def check(self, printer=print):
        """
        Empty debugging check to implement on processors.
        """
        return []



from .base import ComposerProcessor
from .classes import ClassProcessor
from .purge import PurgeProcessor
from .text import TextContentProcessor


__all__ = [
    "ComposerProcessor",
    "ClassProcessor",
    "PurgeProcessor",
    "TextContentProcessor",
]



import shutil
from pathlib import Path

from ..importer import import_module

from .base import ComposerProcessor
from ..exceptions import ComposerPurgeError


class PurgeProcessor(ComposerProcessor):
    """
    Processor to purge application repository from not enabled application.

    This processor requires the repository to be set and valid.
    """
    def export(self):
        """
        Export module directory paths that should be purged from repository.

        Returns:
            list: A list of Path objects.
        """
        try:
            repository = import_module(self.composer.manifest.repository)
        except ModuleNotFoundError:
            msg = "{klass} is unable to find application repository module from: {path}"
            raise ComposerPurgeError(msg.format(
                klass=self.__class__.__name__,
                path=self.composer.manifest.repository,
            ))

        repository_path = Path(repository.__file__).parent

        # List module directories from repository and filter out the ones with a name
        # starting with "_"
        appdirs = [
            child
            for child in repository_path.iterdir()
            if (child.is_dir() and not child.name.startswith("_"))
        ]

        # Filter out the application module directories that are enabled from manifest
        return [
            item
            for item in appdirs
            if item.name not in [a.name for a in self.composer.apps]
        ]

    def commit(self):
        """
        Commit repository purge.
        """
        to_remove = self.export()

        for path in to_remove:
            msg = "{klass} is removing application: {path}".format(
                klass=self.__class__.__name__,
                path=path,
            )
            self.composer.log.info(msg)
            shutil.rmtree(path)

        return to_remove



import json


def debug_invoke(result, caplog=None):
    """
    Output important informations on Click invoke result and optionally caplog.

    It invoke command has raised an exception, raise it up.
    """
    print("=> result.output <=")
    print(result.output)
    print()

    if caplog:
        print("=> caplog.record_tuples <=")
        print(caplog.record_tuples)
        print()

    print("=> result.exception <=")
    print(result.exception)
    if result.exception is not None:
        raise result.exception


def dump_datasets(basedir, source_name, source, result, resolved_payload):
    """
    A helper function to dump source and resolving result lists in JSON files.
    """
    apps_filename = result.replace("_result", "_apps")

    source_destination = basedir / source_name
    result_destination = basedir / result
    apps_destination = basedir / apps_filename

    print("source_destination:", source_destination)
    print("result_destination:", result_destination)
    print("apps_destination:", apps_destination)

    # Write the collection source
    # print("SRC-"*30)
    # print()
    # print(source)
    source_destination.write_text(source)

    # Write the resolved collection as payload items
    result_content = json.dumps(resolved_payload, indent=4)
    # print("PAYLOAD-"*20)
    # print()
    # print(result_content)
    result_destination.write_text(result_content)

    # Write the resolved collection as just application names
    apps = json.dumps([item.get("name") for item in resolved_payload], indent=4)
    # print("APPS-"*30)
    # print()
    # print(apps)
    apps_destination.write_text(apps)

    return source_destination, result_destination, apps_destination
